from scrapeanything.scraper.scraper import Scraper

import requests
import lxml.html
from lxml import etree

class Requests(Scraper):

    def on_wget(self, url):
        headers = { 'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36' }
        page = requests.get(url, headers=headers, verify=True, proxies=None)
        if page.content == b'':
            raise Exception()

        return lxml.html.fromstring(page.content)

    def on_xPath(self, element, path, prop=None):
        return element.xpath(path)

    def on_get_text(self, element):
        return element.text_content()

    def on_get_html(self, element):
        return etree.tostring(element, pretty_print=True)

    def on_get_attribute(self, element, prop):
        return element.get(prop)

    def on_get_css(self, element, prop):
        pass

    def on_close(self):
        pass


from contextlib import contextmanager
from scrapeanything.utils.config import Config
from scrapeanything.database.repository import Repository

@contextmanager
def Connection(config: Config):

    repository = Repository(config=config)
    try:
        repository.connect()
        yield repository
        repository.commit()
    except:
        repository.rollback()
    finally:
        repository.close()


from scrapeanything.utils.config import Config
from scrapeanything.database.repository import Repository
from scrapeanything.integration.excel import Excel
from scrapeanything.integration.csv import Csv
from scrapeanything.scraper.parser import Parser
from scrapeanything.database.models import Model
from scrapeanything.utils.constants import *

import requests
from enum import Enum

class Methods(Enum):
    POST = 'POST'
    GET = 'GET'
    PUT = 'PUT'

class Service:

    def __init__(self, config: Config=None, repository: Repository=None) -> None:
        if repository is not None:
            self.repository = repository
    
        if config is not None:
            self.config = config
    
    def __del__(self):
        if hasattr(self, 'repository'):
            self.repository.close()

    def all(self) -> 'list[Model]':
        entities = self.repository.all(entity_type=self.entity_type)
        return entities

    def load(self, entity: Model) -> Model:
        entity = self.repository.load(entity=entity)
        return entity

    def save(self, entity: Model) -> None:
        self.repository.save(entity=entity)

    def wget(self, url: str, parameters: dict=None, method: Methods=Methods.GET) -> any:
        return requests.get(url=url, data=parameters).text



from datetime import datetime
from decimal import Decimal
import sys
import re
from time import strptime
import unicodedata
from money import Money

class Types:
    @staticmethod
    def is_primitive(value: any) -> bool:
        primitive = (int, float, str, bool, datetime)
        return type(value) in primitive

    @staticmethod
    def to_class(module_name: str, class_name: str) -> any:
        return getattr(sys.modules[module_name], class_name)

    @staticmethod
    def to_date(value: str, format: str) -> datetime:
        return strptime(string=value, format=format)

    @staticmethod
    def to_number(value: any) -> float:
        if value is None:
            return None
        else:
            if bool(re.search(r'\d', value)) == True:
                value = ''.join(i for i in value if i.isdigit() or i in [ ',', '.', '-' ])
                value = value[0] + value[1:].replace('-', '', 999) # if it is a negative number, remove all '-' except the first '-' (i.e. -10)
                value = value.replace(',', '.')
                return float(value)
            else:
                return None

    @staticmethod
    def to_integer(value: any) -> int:
        value = Types.to_number(value)
        return int(value)

    @staticmethod
    def to_currency(value: any) -> str:
        value = Money(amount=value, currency='EUR')
        return str(value)

    @staticmethod
    def to_geo(value: any) -> int:
        if value is not None:
            return value.replace(',', '')
        else:
            return None

    @staticmethod
    def to_money(value: any) -> Decimal:
        if value is not None:
            thousand_seaparator = '.'
            decimal_separator = ','
            currency_sign = '€'
            value = value.replace(unicodedata.lookup(currency_sign), '').replace(thousand_seaparator, '').replace(decimal_separator, '.').strip()

            values = value.split(' ')
            value = None
            for v in values:
                if Types.to_number(v) is not None:
                    value = v               

            if value is not None and value.isnumeric():
                return Decimal(value)
            else:
                return None
        else:
            return None

    @staticmethod
    def to_char(value: any) -> str:
        if value is None:
            return None
        else:
            return value.strip()

    @staticmethod
    def to_list(klist: list) -> list:
        return [item for sublist in klist for item in sublist]

    @staticmethod
    def to_dict(obj: any, columns: list) -> dict:
        dictionary = {}
        for attr in dir(obj):
            if attr in columns:
                dictionary[attr] = getattr(obj, attr)

        return dictionary


from scrapeanything.scraper.scraper import Scraper

from selenium import webdriver
from selenium.webdriver.common.keys import Keys
from selenium import webdriver
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import Select
from selenium.common.exceptions import NoSuchElementException        
from selenium.common.exceptions import StaleElementReferenceException

from selenium.webdriver.remote.remote_connection import LOGGER

import logging
import platform
import base64

class Selenium(Scraper):

    def __init__(self, headless=True):
        options = webdriver.ChromeOptions()
        
        options.add_argument('--no-sandbox')
        options.add_argument('--log-level=3')
        options.add_argument('start-maximized')
        options.add_argument('disable-infobars')
        options.add_argument('--disable-extensions')
        options.add_argument('--disable-dev-shm-usage')

        if headless is True:
            options.add_argument('headless')

        LOGGER.setLevel(logging.WARNING)

        if platform.system() == 'Windows':
            self.driver = webdriver.Chrome(executable_path='c:\chromedriver.exe', options=options)
        if platform.system() == 'Linux':
            self.driver = webdriver.Chrome(executable_path='/chromedriver', options=options)
        
        self.handles = {}

        super().__init__()

    def on_wget(self, url):
        self.driver.get(url)
        return self.driver

    def on_xPath(self, path, element=None, timeout=0):

        try:
            el = self.on_exists(element=element, path=path, timeout=timeout)
            if el != False:
                return (self.driver if element is None else element).find_elements(by=By.XPATH, value=path)
            else:
                return None
        
        except NoSuchElementException:
            return None

        return element

    def on_exists(self, path, element, timeout=0):
        element = self.driver if element is None else element

        if timeout > 0:
            try:
                return WebDriverWait(element, timeout).until(
                    EC.presence_of_element_located((By.XPATH, path))
                )
            except Exception as e:
                return False
        else:
            try:
                return element.find_elements(by=By.XPATH, value=path)

            except NoSuchElementException:
                return False
            except StaleElementReferenceException:
                raise StaleElementReferenceException()

    def on_get_text(self, element):
        return element.text

    def on_get_html(self, element):
        return element.get_attribute('innerHTML')

    def on_get_attribute(self, element, prop):
        return element.get_attribute(prop)

    def on_click(self, path, element=None, timeout=0):
        element = self.on_xPath(path=path, element=element, timeout=timeout)

        if element is not None:
            if type(element) is list:
                element = element[0]

            try:
                self.driver.execute_script("arguments[0].click();", element)
            except Exception as e:
                element.click()                

            return element
        else:
            return None

    def on_back(self):
        self.driver.back()

    def on_get_current_url(self):
        return self.driver.current_url

    def on_enter_text(self, path, text, clear=False, element=None):
        element = self.on_xPath(path, element)
        if element is not None:
            if type(element) is list:
                element = element[0]

            if clear is True:
                element.send_keys(Keys.CONTROL + 'a')
                element.send_keys(Keys.BACKSPACE)

            element.send_keys(text)
            return True
        else:
            return False

    def on_solve_captcha(self, path):
        pass

    def on_login(self, username_text=None, username=None, password_text=None, password=None):
        self.on_enter_text(username_text, username)
        self.on_enter_text(password_text, password)

        self.on_click('//button[@class="sign-in-form__submit-button"]')

    def on_search(self, path=None, text=None):
        self.on_enter_text(path, text)
        return self.on_enter_text(path, Keys.RETURN)

    def on_scroll_to_bottom(self):
        self.driver.execute_script('window.scrollTo(0, document.body.scrollHeight);')

    def on_get_scroll_top(self):
        return self.driver.execute_script('return document.documentElement.scrollTop;')

    def on_get_scroll_bottom(self):
        return self.driver.execute_script('return document.body.scrollHeight')

    def on_select(self, path: str, option: str, element: any=None) -> None:
        element = WebDriverWait(self.driver if element is None else element, 10).until(
            EC.presence_of_element_located((By.XPATH, path))
        )

        if element is not None:
            Select(element).select_by_visible_text(option)

    def on_get_image_from_canvas(self, path, local_path, element):
        image_filename = f'{local_path}/image.png'

        # get the base64 representation of the canvas image (the part substring(21) is for removing the padding "data:image/png;base64")
        base64_image = self.driver.execute_script(f'return document.querySelector("{path}").toDataURL("image/png").substring(21);')

        # decode the base64 image
        output_image = base64.b64decode(base64_image)

        # save to the output image
        with open(image_filename, 'wb') as f:
            f.write(output_image)

        return image_filename

    def on_close(self):
        self.driver.quit()


import os
from openpyxl import load_workbook
import pandas as pd

class Excel:

    def __init__(self):
        pass

    def read_from_excel(self, filename, sheet_name='Sheet1', header=0, index_col=False, keep_default_na=True, skiprows=None):
        return pd.read_excel(filename, sheet_name=sheet_name, header=header, index_col=None, keep_default_na=keep_default_na, skiprows=skiprows, engine='openpyxl', parse_dates=[0])

    def append_to_excel(self, filename, data, sheet_name='Sheet1', startrow=None, truncate_sheet=False, header=False, **to_excel_kwargs):
        """
        Append a DataFrame [df] to existing Excel file [filename]
        into [sheet_name] Sheet.
        If [filename] doesn't exist, then this function will create it.

        @param filename: File path or existing ExcelWriter
                        (Example: '/path/to/file.xlsx')
        @param df: DataFrame to save to workbook
        @param sheet_name: Name of sheet which will contain DataFrame.
                        (default: 'Sheet1')
        @param startrow: upper left cell row to dump data frame.
                        Per default (startrow=None) calculate the last row
                        in the existing DF and write to the next row...
        @param truncate_sheet: truncate (remove and recreate) [sheet_name]
                            before writing DataFrame to Excel file
        @param to_excel_kwargs: arguments which will be passed to `DataFrame.to_excel()`
                                [can be a dictionary]
        @return: None

        Usage examples:

        >>> append_df_to_excel('d:/temp/test.xlsx', df)

        >>> append_df_to_excel('d:/temp/test.xlsx', df, header=None, index=False)

        >>> append_df_to_excel('d:/temp/test.xlsx', df, sheet_name='Sheet2',
                            index=False)

        >>> append_df_to_excel('d:/temp/test.xlsx', df, sheet_name='Sheet2', 
                            index=False, startrow=25)

        (c) [MaxU](https://stackoverflow.com/users/5741205/maxu?tab=profile)
        """

        df = pd.DataFrame(data)

        # Excel file doesn't exist - saving and exiting
        if not os.path.isfile(filename):
            
            df.to_excel(
                filename,
                sheet_name=sheet_name, 
                startrow=startrow if startrow is not None else 0, 
                **to_excel_kwargs)
            return
        
        # ignore [engine] parameter if it was passed
        if 'engine' in to_excel_kwargs:
            to_excel_kwargs.pop('engine')

        writer = pd.ExcelWriter(filename, engine='openpyxl', mode='a')

        # try to open an existing workbook
        writer.book = load_workbook(filename)

        if sheet_name in writer.book.sheetnames:
            header = False
        
        # get the last row in the existing Excel sheet
        # if it was not specified explicitly
        if startrow is None and sheet_name in writer.book.sheetnames:
            startrow = writer.book[sheet_name].max_row

        # truncate sheet
        if truncate_sheet and sheet_name in writer.book.sheetnames:
            # index of [sheet_name] sheet
            idx = writer.book.sheetnames.index(sheet_name)
            # remove [sheet_name]
            writer.book.remove(writer.book.worksheets[idx])
            # create an empty sheet [sheet_name] using old index
            writer.book.create_sheet(sheet_name, idx)

        # copy new sheet from template sheet
        if 'template' in writer.book.sheetnames and sheet_name not in writer.book.sheetnames:
            sheet_to_copy = writer.book.get_sheet_by_name('template')
            sheet_copied = writer.book.copy_worksheet(sheet_to_copy)
            sheet_copied.title = sheet_name

        # copy existing sheets
        writer.sheets = { ws.title:ws for ws in writer.book.worksheets }

        if startrow is None:
            startrow = 0

        # make columns filterable
        writer.book.active.auto_filter.ref = writer.book.active.dimensions

        # make columns autofit
        # TODO

        # write out the new sheet
        df.to_excel(writer, sheet_name, startrow=startrow, header=header, **to_excel_kwargs)

        # save the workbook
        writer.save()

    def write_to_excel(self, filename, data, sheet_name='Sheet1'):
        # create excel writer object
        writer = pd.ExcelWriter(filename, datetime_format='hh:mm:ss')

        if type(data) is dict:     
            for sheet_name, dataframe in data.items():
                dataframe.to_excel(writer, sheet_name=sheet_name, index=False)
        else:
            # write dataframe to excel
            data = pd.DataFrame(data)
            data.to_excel(writer, sheet_name=sheet_name, index=False)
        # save the excel
        writer.save()

    def get_sheets(self, filename):
        writer = pd.ExcelWriter(filename, engine='openpyxl', mode='a')

        # try to open an existing workbook
        writer.book = load_workbook(filename)

        return writer.book.sheetnames